# pip install github3.py
import github3
import pandas as pd
import sys, os
import subprocess as sp
import json

def init():
    git_config = sp.run(["git","config","-l"], capture_output=True).stdout
    split_config = git_config.decode().split("\n")
    git_dict = {s[0]:s[1] for s in [ s.split('=') for s in split_config[:-11] ]}

    gh_username = git_dict['user.email']
    gh_password = git_dict['user.token']

    return gh_username, gh_password

def find_forks(repo_url: str):
    gh_username, gh_password = init()

    user, repo = repo_url.split('/')[-2:]

    g = github3.login(gh_username, gh_password)
    r = g.repository(user, repo)
    total_list = list(r.forks(sort='commits', number=r.fork_count))
    data = [ (i,i.updated_at,
            i.pushed_at, i.updated_at == i.created_at) for i in total_list ]
    df = pd.DataFrame(data,columns=['name','updated_at','pushed_at','ever_changed'])

    # TODO: humanize dates/times
    # https://github.com/jmoiron/humanize
    df = df.sort_values('ever_changed')
    print(df.ever_changed.value_counts())
    return df

def fork_status(repo_url: str):
    cmd = 'for branch in `git branch -r | grep -v HEAD`;do echo -e `git show --format="%ci %cr" $branch | head -n 1` \\t$branch; done | sort -r\n'
    os.system(cmd)

def find_contributors(repo_url: str):
    from .sshconf import SshConfig
    # gh_username, gh_password = init()

    user, repo = repo_url.split('/')[-2:]
    f = open('/Users/jakub/.gitconfig').read().splitlines()
    c = SshConfig(f)

    g = github3.login(gh_username, gh_password)
    r = g.repository(user, repo) 

def dir_diff(cwd: str, action = 'status') -> json:
    # get all dirs in cwd
    output = []
    all_directories = os.listdir(cwd)

    # check that all are directories
    dirs_only = [ d for d in all_directories if len(d.split('.'))==1 and d!='swap2' ]

    for dir in dirs_only:
        cwd = os.path.abspath(dir)
        try:
            git_stash = sp.run(["git",action], cwd=cwd, capture_output=True)
        except NotADirectoryError:
            print('Not a directory:', cwd)
        output.append([cwd, git_stash.stdout.decode()])

    diff_df = pd.DataFrame(output, columns=['dir',action])
    if action == 'status':
        # remove if up to master
        diff_df = diff_df[~diff_df.status.str.contains('On branch master')]
        
        # remove if main
        diff_df = diff_df[~diff_df.status.str.contains('On branch main')]
        
        # drop rows with empty ('') status
        diff_df = diff_df[diff_df.status != '']
        
        # create column up to date with True/False
        diff_df['up_to_date'] = diff_df.status.str.contains('up to date')
        diff_df = diff_df.sort_values(by='up_to_date')
        diff_df.reset_index(inplace=True)
        
        return diff_df
        
    else:
        non_empty_diffs = diff_df.loc[diff_df['diff'].str.len() > 0]
        dict_diffs = non_empty_diffs.to_dict(orient='index')

        # map 'dir' value to key in dict_diffs
        dict_diffs = {v['dir']:v['diff'] for k,v in dict_diffs.items()}

    return dict_diffs


#/bin/python
'''
TODO: fix

botocore.exceptions.ClientError: An error occurred (InvalidPermission.Duplicate) 
when calling the AuthorizeSecurityGroupIngress operation: the specified rule "peer: 86.16.93.195/32, TCP, from port: 8501, 
to port: 8501, ALLOW" already exists

  File "/Users/jlangr/code/data-toolkit/dt/controllers/base.py", line 101, in sg
    update_sg_to_ip(sg_name=sg_name, profile=profile,
  File "/Users/jlangr/code/data-toolkit/dt/ext/aws_sg.py", line 111, in update_sg_to_ip
    data = ec2.authorize_security_group_ingress(
'''

import boto3
import pandas as pd
import os 

try:
    term_wdith = os.get_terminal_size().columns
except OSError:
    term_wdith = 70

# pd.set_option('display.max_colwidth',term_wdith - 40)
pd.set_option('display.max_colwidth',30)

def get_public_ip():
    from urllib.request import urlopen
    import re
    data = str(urlopen('http://checkip.dyndns.com/').read())
    # data = '<html><head><title>Current IP Check</title></head><body>Current IP Address: 65.96.168.198</body></html>\r\n'
    return re.compile(r'Address: (\d+\.\d+\.\d+\.\d+)').search(data).group(1)

def get_sg(profile: str, sid: int, sg_name = None, region='eu-west-1'):
    
    # decide if use security group name or security group id
    if sid:
        # get relevant row from security group df
        sg_name = list_security_groups(profile=profile, region=region).iloc[sid]['GroupName']
    else:
        sg_name = 'default'
    
    ec2 = boto3.session.Session(profile_name=profile, region_name=region).client('ec2')
    resp = ec2.describe_security_groups()
    target_sg_dict = list(filter(lambda x: x['GroupName']==sg_name, resp['SecurityGroups']))
    return target_sg_dict

def check_if_exists(ip_ranges: list, descriptions: list):
    # for each description check if there is a description in the existing permissions with the same description
    try:
        # TODO: why is 'ip_ranges' not defined?
        return any([ ip_ranges['Description']==d for d in descriptions ])
        # *** NameError: name 'ip_ranges' is not defined
        # return 'Aux' in ip_ranges['IpRanges'][0]['Description']
    except IndexError as e:
        assert ip_ranges['IpRanges']==[]
        return False
    except KeyError as e:
        return False

def update_sg_to_ip(sg_name='jakub', # ip='2.222.105.71/32',
                    ports=[22, 8888, 8501, 8889, 5432], profile='default',
                    messages='', region='eu-west-1'):
    import boto3

    ip = f"{get_public_ip()}/32"

    try:
        group_name = list_security_groups(profile=profile, region=region)
        sid = group_name.index[group_name.GroupName==sg_name].tolist()[0]
    except IndexError:
        raise ValueError(f"Security group called {sg_name} does not exist")
    ec2 = boto3.session.Session(profile_name=profile, region_name=region).client('ec2')

    target_sg_dict = get_sg(sg_name=sg_name, sid=sid, profile=profile, region=region)[0]
    ec2_res = boto3.session.Session(profile_name=profile, region_name=region).resource('ec2')
    sg_res = ec2_res.SecurityGroup(target_sg_dict['GroupId'])

    # determine what the new sg rules should be either default or custom
    if messages=='':
        print('Warning messages are empty. Skipping')
        descriptions = {
                22: 'Jakub Aux SSH',
                80: 'Web',
                8888: 'Jakub Aux Jupyter',
                8501: 'Jakub Aux Streamlit',
                8889: 'Jakub Aux Jupyter2',
                5432: 'Jakub Aux Postgres'
            }
    else:
        items = [ i for i in messages.split(',')]
        descriptions = { int(k.split(':')[0]):k.split(':')[1] for k in items } 
        ports = [ int(i.split(':')[0]) for i in items ] 

    # TODO: allow for more than just 1 port setting
    target_ports = [ [ s['ToPort']==p for p in ports ]
                    for s in sg_res.ip_permissions if 'ToPort' in s.keys()]
    # flatten target_ports list take all True values and merge into one list
    # [[True, False, False],[False, False True]] -> [True, False, True]
    target_ports = [ any(i) for i in zip(*target_ports) ]
     
    # loop using description to only remove already existing and replace them later
    # effectiely overwriting them
    permissions_to_revoke = []
    for i in range(len(sg_res.ip_permissions)):
        if target_ports[i]:
            ipr = sg_res.ip_permissions[i]
            ip_ranges_to_revoke = [ s for s in ipr['IpRanges'] if check_if_exists(s, list(descriptions.values())) ]
            permissions_to_revoke.append({ 
                **ipr,
                'IpRanges': ip_ranges_to_revoke })

    # Removes all relevant permissions from this group (originally as `jakub`)
    if sg_res.ip_permissions!=[] and permissions_to_revoke==[]:
        print('No permissions to revoke. Skipping.')
    else:
        sg_res.revoke_ingress(IpPermissions=permissions_to_revoke)
        print(f'Removed permissions {permissions_to_revoke}')

    # assign new rules   
    ip_perms = [ {
        'IpProtocol': 'tcp',
        'FromPort': ports[i],
        'ToPort': ports[i],
        'IpRanges': [{'CidrIp': ip, 'Description': descriptions[ports[i]] }]}
            for i in range(len(ports)) ]
    
    data = ec2.authorize_security_group_ingress(
            GroupId=target_sg_dict['GroupId'],
            IpPermissions=ip_perms )
    print(f'Ingress Successfully Set {ip_perms} to {sg_name}')

# lists all security groups
def list_security_groups(region: str = 'eu-west-1', profile: str = 'default'):
    # define a boto3 client with a region eu-est
    boto3.setup_default_session(profile_name=profile)
    ec2 = boto3.client('ec2', region_name=region)
    response = ec2.describe_security_groups()
    security_group_df = pd.DataFrame(response['SecurityGroups'])

    # extract keys from Tags column
    # [ x for x in df.Tags if not np.all(pd.isna(x)) ] 

    # generate a mask for terms that have FILTER_TERMS in them
    FILTER_TERMS = ['LIW']
    mask = security_group_df.Description.apply(lambda x: any(term in x for term in FILTER_TERMS))
    security_group_df = security_group_df[~mask]

    # display all columns except FILTER_COLS
    FILTER_COLS = ['VpcId', 'Tags','OwnerId']
    security_group_df = security_group_df[security_group_df.columns.difference(FILTER_COLS)]

    # create an Ingress dataframe by filtering on IpPermissions not empty
    df_ingress = security_group_df[security_group_df.IpPermissions.notnull()]
    ip_permissions_df = pd.DataFrame.from_dict([x[0] for x in df_ingress.IpPermissions.values])
    
    df_ingress = df_ingress.reset_index().drop('index',axis=1)
    
    viewable_df = pd.concat([ip_permissions_df,df_ingress],axis=1).drop(['Ipv6Ranges','IpPermissions'],axis=1)
    # viewable_df.Description = viewable_df.Description.str[:30]

    # reorders columns 
    order = "GroupName FromPort IpProtocol ToPort UserIdGroupPairs PrefixListIds Description IpRanges IpPermissionsEgress GroupId".split()
    viewable_df = viewable_df[order]

    return viewable_df




import os
import requests
import json
import time

# construct a function that loads my openai API key from home and then calls it
def load_api_key(api_key_file):
    api_key = json.loads(open(api_key_file).read())['cai_key']
    return api_key

# construct a function that calls the openai codex and returns the results
def constuct_call(api_key, prompt):
    engine = 'cushman-codex' # codex engine could also be
    # engine = 'davinci-codex'
    url = f'https://api.openai.com/v1/engines/{engine}/completions'
    data = { 'prompt': prompt, 'max_tokens' : 200  } # 'stream': True

    # convert data to json
    json_data = json.dumps(data)

    headers = {'Authorization': 'Bearer {}'.format(api_key),
                'Content-Type': 'application/json'}
    response = requests.post(url, json_data, headers=headers)
    return response
    
def get_cai_suggestions(query: str, n_suggestions: int = 5):
    # measure timing 
    start = time.time()
    # get the api key from the home directory 
    api_path = os.path.join(os.path.expanduser('~'), 'dt_config.json')
    api_key = load_api_key(api_path)
    response = constuct_call(api_key, query)
    after = time.time()
    print('API call took {} seconds'.format(after - start))
    # convert response to dict from string 
    response_dict = json.loads(response.text)

    end = time.time()
    print('Total time was {} seconds'.format(end - start))

    return response_dict




import sys
import os
import re
import requests
import shutil
import tempfile
import subprocess
import tqdm

def download_file_from_google_drive(id, destination):
    URL = "https://docs.google.com/uc?export=download"

    session = requests.Session()

    response = session.get(URL, params={'id': id}, stream=True)
    
    # parse response using re to look for html tag id="downloadForm"
    match = re.search(r'id="downloadForm" action="(.+?)"',
                      str(response.content))
    if match:
        target_url = match.group(1)
        # remove "amp;"
        target_url = target_url.replace("amp;", "")
    else:
        target_url = target_url
        
    # download file from destiation using urllib and tqdm
    response = requests.get(target_url, stream=True)
    total_size_in_bytes= int(response.headers.get('content-length', 0))
    block_size = 1024 #1 Kibibyte
    progress_bar = tqdm.tqdm(total=total_size_in_bytes, unit='iB', unit_scale=True)
    with requests.get(target_url, stream=True) as response:
        with open(destination,'wb') as f:
            for data in response.iter_content(block_size):
                progress_bar.update(len(data))
                f.write(data)
    progress_bar.close()
    if total_size_in_bytes != 0 and progress_bar.n != total_size_in_bytes:
        print("ERROR, something went wrong")


import streamlit as st
import pandas as pd
import numpy as np
import os    
import matplotlib.pyplot as plt
from PIL import Image
import random
import fire
    
try:
    import cv2
except ModuleNotFoundError:
    import os
    os.system('pip install streamlit pandas fire matplotlib numpy pillow')
    
# IM_TYPE = ['rec_A', 'fake_B','real_A']
IM_TYPE = ['fake_A','rec_B','real_B']

    
def get_full_img_path(file_path, image_name, image_type):
    # file_path = file_path.replace(IM_TYPE[0],f"{image_type}")
    # image_str = image_name.replace(IM_TYPE[1],f"{image_type}")\
    image_str = f"{image_name.split('_')[0]}_{image_type}.png"
    file_path = f"{'/'.join(file_path.split('/')[:-1])}/{image_type}"
    
    adapted_img = f"{file_path}/{image_str}" 
    return adapted_img


def process_one(image_type: str, image_name: str, percentage: float,
                image_root: str, adapted_root: str):
    # TODO: genericise. Currently assumes all file names are the same.
    # target_image = f'{image_root}/{image_type}/{"_".join(image_name.split("_"))}'
    target_image = get_full_img_path(image_root, image_name, IM_TYPE[-1])
                                    #  image_type.replace('A','B'))
    
    img = cv2.imread(target_image)[:,:,::-1]
    img = cv2.resize(img, (512,512))

    adapted_img = get_full_img_path(adapted_root, image_name, image_type)
    adapted_img = np.array(Image.open(adapted_img).resize((512,512)))
    # adapted_img = cv2.imread(adapted_img)
    adapted_img.resize((512,512,3))
    # adapted_img = adapted_img[:,:,::-1]

    img_adj = int( img.shape[1] * percentage )
    reverse_adj = img.shape[1] - img_adj
    result_img = np.hstack(( img[:,:img_adj,:], adapted_img[:, img_adj:,:] ))
    # result_img = np.concat([img[:img_adj], adapted_img[img_adj:]])
    return result_img



def create_viz(
    images = '/efs/public_data/gta5/images',
    labels = 'labels',
    adapted = '/efs/public_data/images'):
    
    labels = '/'.join(images.split('/')[:-1]) + labels
    # set up ui
    label_paths = sorted([ x for x in os.listdir(labels) ])[:1000][::-1]
    image_paths = sorted([ x for x in os.listdir(images) ])[:1000][::-1]
    adapted_paths = sorted([ x for x in os.listdir(adapted) ])[2:1000][::-1]

    tar_im_type = st.sidebar.selectbox("Which type of image?", IM_TYPE)

    file_selection = st.sidebar.selectbox("Which file?", adapted_paths)

    percentage = st.sidebar.slider("percentage", 0, 100) / 100

    # if st.button('Random Image'):
    # file_selection = image_paths[0]
    #     i = image_paths.index(file_selection)
    #     file_selection = image_paths[i+var_i]
    #     var_i += 1
    st.code(f"{adapted}/{file_selection}")
    # else:
    #     st.write(f'Offset {var_i}')

    # image_root = '/'.join(images.split('/')[:-1])
    # adapted_root = '/'.join(adapted.split('/')[:-1])

    col1, col2, col3 = st.columns(3)
    
    col1.header("Original")
    kpd_img_2 = process_one(IM_TYPE[1], file_selection, 1,
                        images, adapted)
    col1.image(kpd_img_2, use_column_width=True)
    
    col2.header("Mix")
    kpd_img = process_one(tar_im_type, file_selection, percentage,
                        images, adapted)
    col2.image(kpd_img, use_column_width=True)

    col3.header("Adapted")
    kpd_img_1 = process_one(tar_im_type, file_selection, 0,
                        images, adapted)
    col3.image(kpd_img_1, use_column_width=True)



# create_viz(images = '/efs/public_data/gta5/images',
#     labels = 'labels',
#     adapted = '/efs/public_data/images')
if __name__ == "__main__":
    fire.Fire(create_viz)


import os

LEVELS = 2

def old_tree(root_path: str, depth: int = 0) -> None:
    """
    Prints a tree of directories and files.
    """
    # exclude hidden files
    files_list = os.listdir(root_path)
    files_list = [f for f in files_list if not f.startswith('.')]

    # print the directories
    for dir in os.listdir(root_path):
        if os.path.isdir(os.path.join(root_path, dir)):
            tree(os.path.join(root_path, dir), depth + 1)
    for dir_name in files_list:
        dir_path = os.path.join(root_path, dir_name)
        if os.path.isdir(dir_path):
            # full file_name (path)
            file_name = os.path.join(root_path, dir_name)
            if not file_name.startswith('.'):
                print(' ' * depth * LEVELS , file_name)
                tree(dir_path, depth + 1)
        else:
            file_name = os.path.join(root_path, dir_name)
            if not file_name.startswith('./.git'):
                print(' ' * depth * LEVELS, file_name)
                
                
def tree(path,show_files=False,full_path=0, indentation=2,file_output=False):
    """
    Shows the content of a folder in a tree structure.
    path -(string)- path of the root folder we want to show.
    show_files -(boolean)-  Whether or not we want to see files listed.
                            Defaults to False.
    indentation -(int)- Indentation we want to use, defaults to 2.   
    file_output -(string)-  Path (including the name) of the file where we want
                            to save the tree.
    """
    tree = []
    
    if not show_files:
        for root, dirs, files in os.walk(path):
            if not root.startswith('./.'): # TODO: parametrize this
                level = root.replace(path, '').count(os.sep)
                indent = ' '*indentation*(level)
                tree.append('{}{}/'.format(indent,os.path.basename(root)))

    if show_files:
        for root, dirs, files in os.walk(path):
            if not root.startswith('./.'): # TODO: parametrize this
                level = root.replace(path, '').count(os.sep)
                indent = ' '*indentation*(level)
                tree.append('{}{}/'.format(indent,os.path.basename(root)))    
                for f in files:
                    if full_path: f = os.path.join(root, f)
                    subindent=' ' * indentation * (level+1)
                    tree.append('{}{}'.format(subindent,f))

    if file_output:
        output_file = open(file_output,'w')
        for line in tree:
            output_file.write(line)
            output_file.write('\n')
    else:
        # Default behaviour: print on screen.
        for line in tree:
            print(line)


import json
import os
import humanize
import time

# construct a function that loads my openai API key from home and then calls it
def load_config():
    config_path = os.path.join(os.path.expanduser('~'), 'dt_config.json')
    config = json.loads(open(config_path).read())   
    return config


def show_config():
    config = load_config()
    
    for k in config.keys():
        # specific printing for snippets
        if isinstance(config[k], list):
            print(f"{k}")
            for s in config[k]:
                # gets original stringified time and converts into timedelta from now
                orig_timestamp = time.strptime(s[1], '%Y-%m-%d %H-%M-%S')
                delta = humanize.naturaltime(time.time() - time.mktime(orig_timestamp))
                print(f" {s[0]} ({delta}): \n  {s[2]}")   
        else:
            print(f"{k}: \n {config[k]}")




# MIT Licensed code
import os
import boto3
from collections import namedtuple
from operator import attrgetter


S3Obj = namedtuple('S3Obj', ['key', 'mtime', 'size', 'ETag'])


def s3list(bucket, path, start=None, end=None, recursive=True, list_dirs=True,
           list_objs=True, limit=None):
    """
    Iterator that lists a bucket's objects under path, (optionally) starting with
    start and ending before end.

    If recursive is False, then list only the "depth=0" items (dirs and objects).

    If recursive is True, then list recursively all objects (no dirs).

    Args:
        bucket:
            a boto3.resource('s3').Bucket().
        path:
            a directory in the bucket.
        start:
            optional: start key, inclusive (may be a relative path under path, or
            absolute in the bucket)
        end:
            optional: stop key, exclusive (may be a relative path under path, or
            absolute in the bucket)
        recursive:
            optional, default True. If True, lists only objects. If False, lists
            only depth 0 "directories" and objects.
        list_dirs:
            optional, default True. Has no effect in recursive listing. On
            non-recursive listing, if False, then directories are omitted.
        list_objs:
            optional, default True. If False, then directories are omitted.
        limit:
            optional. If specified, then lists at most this many items.

    Returns:
        an iterator of S3Obj.

    Examples:
        # set up
        >>> s3 = boto3.resource('s3')
        ... bucket = s3.Bucket('bucket-name')

        # iterate through all S3 objects under some dir
        >>> for p in s3list(bucket, 'some/dir'):
        ...     print(p)

        # iterate through up to 20 S3 objects under some dir, starting with foo_0010
        >>> for p in s3list(bucket, 'some/dir', limit=20, start='foo_0010'):
        ...     print(p)

        # non-recursive listing under some dir:
        >>> for p in s3list(bucket, 'some/dir', recursive=False):
        ...     print(p)

        # non-recursive listing under some dir, listing only dirs:
        >>> for p in s3list(bucket, 'some/dir', recursive=False, list_objs=False):
        ...     print(p)
"""
    kwargs = dict()
    if start is not None:
        if not start.startswith(path):
            start = os.path.join(path, start)
        # note: need to use a string just smaller than start, because
        # the list_object API specifies that start is excluded (the first
        # result is *after* start).
        kwargs.update(Marker=__prev_str(start))
    if end is not None:
        if not end.startswith(path):
            end = os.path.join(path, end)
    if not recursive:
        kwargs.update(Delimiter='/')
        if not path.endswith('/'):
            path += '/'
    kwargs.update(Prefix=path)
    if limit is not None:
        kwargs.update(PaginationConfig={'MaxItems': limit})

    paginator = bucket.meta.client.get_paginator('list_objects')
    for resp in paginator.paginate(Bucket=bucket.name, **kwargs):
        q = []
        if 'CommonPrefixes' in resp and list_dirs:
            q = [S3Obj(f['Prefix'], None, None, None) for f in resp['CommonPrefixes']]
        if 'Contents' in resp and list_objs:
            q += [S3Obj(f['Key'], f['LastModified'], f['Size'], f['ETag']) for f in resp['Contents']]
        # note: even with sorted lists, it is faster to sort(a+b)
        # than heapq.merge(a, b) at least up to 10K elements in each list
        q = sorted(q, key=attrgetter('key'))
        if limit is not None:
            q = q[:limit]
            limit -= len(q)
        for p in q:
            if end is not None and p.key >= end:
                return
            yield p


def __prev_str(s):
    if len(s) == 0:
        return s
    s, c = s[:-1], ord(s[-1])
    if c > 0:
        s += chr(c - 1)
    s += ''.join(['\u7FFF' for _ in range(10)])
    return s


from ..ext.hist import hist_tail
import os
import json
import pandas as pd
import subprocess
import pyperclip

def get_command(n_lines: list):
    # TODO: propagate that last line in n_lines is the # of lines used
    # maybe add a check for 1 num to use default (20)
    df = hist_tail(n_lines[-1] * 2)
    exec_line = f"{df.iloc[-1]['Command']}"
    for cmd in n_lines[:-1]:
        exec_line += ' && '
        exec_line += f"{df.iloc[cmd].Command}"
        
    return exec_line
    
def run_command_from_alias(alias: str):
    from dt.ext.config import load_config
    config = load_config()['snippets']
    df = pd.DataFrame(config)
    df.columns = ['Name','Time','Command']
    
    list_to_run = df[df['Name'] == alias]['Command'].values[0]
    cmd_to_run = " && ".join(list_to_run)
    
    # TODO: maybe replace with this fix for source
    # https://stackoverflow.com/questions/7040592/calling-the-source-command-from-subprocess-popen
    print(f"Copied {cmd_to_run} to clipboard")
    pyperclip.copy(cmd_to_run)
    # Maybe we can use this if there are no problematic commands
    # subprocess.Popen(cmd_to_run, shell=True)
    
def run_command(n_lines: list):
    exec_line = get_command(n_lines)
    print(f"Running {exec_line[:-4]}")
    os.system(exec_line[:-4])

    
def save_command(n_lines: list):
    raise NotImplementedError('This function is not deprecated. Please use `dt hist -s`')
    
def list_snippets(ls_level: int):
    snippets_path = os.path.join(os.path.expanduser('~'), 'dt_config.json')
    config = json.loads(open(snippets_path).read())
    print(f"Snippets: {config['snippets']}")


import asana
import json
import os
# import request
import pandas as pd

# pandas set max column width
pd.set_option('display.max_colwidth', 150)


def user_select_option(message, options):
    option_lst = list(options)
    print_(message)
    for i, val in enumerate(option_lst):
        print_(i, ': ' + val['name'])
    index = int(input("Enter choice (default 0): ") or 0)
    return option_lst[index]

def get_client():
    with open(os.path.expanduser('~/.dt_config.json')) as f:
        config = json.load(f)
        api_key = config['asana_api_key']
        
    # create asana client
    client = asana.Client.access_token(api_key)
    return client


def asana_list_todos(filtering, workspace_name):
    # read api key from ~/.dt_config.json
    client = get_client()
    (url, state) = client.session.authorization_url()
        
    me = client.users.me()
    workspace_id = me['workspaces'][0]['gid']
    
    
    # tasks = str(client.tasks)
    tasks = list(client.tasks.find_all(workspace=workspace_id, assignee='me'))
    df = pd.DataFrame(tasks)
    df = df[['name']]
    
    print(df)
    
    
def add_todo(task_text, project_id=0):
    client = get_client()
    me = client.users.me()
    workspace_id = me['workspaces'][0]['gid']
    
    projects = list(client.projects.get_projects_for_workspace(workspace_id))
    
    # docs https://developers.asana.com/docs/create-a-task
    
    data =  {'name': task_text,
        "resource_subtype": "default_task",
        "assignee": me['gid'],
        "due_on": "2019-09-15",
        "projects": projects[project_id]['gid'],
        # 'notes': 'Note: This is a test task created with the python-asana client.',
        # 'projects': [workspace_id]
    }
    
    result = client.tasks.create_in_workspace(workspace_id, data)

    print(json.dumps(result, indent=4))
    
def done_todo():
    pass


'''
Operations to speed up S3 access.
'''
import pandas as pd
import boto3
# from profilehooks import profile
import os
try:
    term_width = os.get_terminal_size().columns
except OSError:
    term_width = 70


DONGLES_TO_SKIP = [
    'test',
    '861579032390149', # not entirely sure why 
]

S3_URL = '' # to be changeable by the user?

def urlifier(dongle_id, drive):
    """
    Return a url for a drive.
    """
    return f'https://{S3_URL}.amazonaws.com/creationlabs-raw-data/{dongle_id}/{drive}'


def get_lastest_file(dongle_id: str):
    """
    Get the lastest file in the bucket.
    TODO: eventually paratemrize the -1 to get any file
    """
    from joblib import delayed, Parallel

    s3_resource = boto3.resource('s3')
    bucket_name = 'creationlabs-raw-data'
    bucket = s3_resource.Bucket(bucket_name)

    last_segment, last_complete_drive = get_selected_drives(None, s3_resource, bucket_name, dongle_id)
    drive = '--'.join(last_segment.drive.split('--')[:-1])
    return urlifier(last_segment.dongle_id, drive)


def get_drives_df(s3_resource, bucket_name, dongle_id: str, filter=0):
    '''
    '''
    from dt.ext.aws_s3_list import s3list

    bucket = s3_resource.Bucket(bucket_name)
    bucket_list = list(s3list(bucket, dongle_id, recursive=False, list_dirs=True))
    df = pd.DataFrame(bucket_list) 

    # filter out files with boot crash or swaglog
    df = df[~df.key.str[-5:-1].isin(['glog','boot','rash'])]

    df = df.key.str.split('/', expand=True)
    try:
        df.columns = ['dongle_id', 'drive','ext']
    except ValueError:
        return pd.DataFrame([])

    df['date'] = df.drive.str[:10]
    df['time'] = df.drive.str[12:20]
    df['seg_num'] = df.drive.str.split('--').str[-1]
    df['drive_time'] = df.drive.str.split('--').str[:-1].str.join('--')

    df = df.sort_values(by=['date','time'], ascending=False)
    df.seg_num = pd.to_numeric(df.seg_num, errors='coerce')

    if filter!=0:
        df = df[df.seg_num!=0]
        df = df.reset_index().groupby('drive_time').max()
        df = df[~pd.isna(df.seg_num)].reset_index()
        
        # set display cols according to terminal width
        pd.set_option('display.max_colwidth',term_width)

    return df

def get_selected_drives(s3, s3_resource, bucket_name, dongle_id):
    """
    Get the selected drives for a dongle.
    """
    df = get_drives_df(s3_resource, bucket_name, dongle_id)
    
    if len(df) == 0:
        # return two empty series of dimensions 14
        return pd.Series(), pd.Series()

    # filter out playlist files/folders
    df = df[df.drive_time.str.len() > 16 ]

    try:
        latest_drive = df.iloc[0]
        last_complete_drives = df[df.seg_num!=0]
    except IndexError:
        return pd.Series(), pd.Series()

    if len(last_complete_drives) > 0:
        last_complete_drive = last_complete_drives.iloc[0]
    else:
        last_complete_drive = latest_drive

    return latest_drive, last_complete_drive

def parallel_get_selected_drives(dongle, bucket_name, latest, latest_files):
    """
    Get the selected drives for a dongle.
    """
    # setting up parameters for parallel processing
    s3 = boto3.session.Session(region_name='eu-west-1').client('s3')
    s3_resource = boto3.resource('s3')
    
    latest_drive, last_complete_drive = get_selected_drives(s3, s3_resource, bucket_name, dongle)
    if latest:       latest_files.append(last_complete_drive)
    else:  latest_files.append(latest_drive)
    return latest_files
    

def get_dongle_drives(s3, s3_resource, bucket_name, dongle_id: str, filter: int, latest=False):
    from pqdm.processes import pqdm
    from functools import partial
    # get number of cores in host computer
    num_cores = os.cpu_count()
   
    raw_data_buckets = s3.list_objects_v2(Bucket='creationlabs-raw-data',Delimiter='/')
    dongle_ids = [x['Prefix'].split('/')[0] for x in raw_data_buckets['CommonPrefixes']]
    
    if filter: dongle_ids = [ x for x in dongle_ids if x not in DONGLES_TO_SKIP]
    else: dongle_ids = [ x for x in dongle_ids if x not in DONGLES_TO_SKIP]

    latest_files = []
    
    # create partial for parallel_get_selected_drives
    parallel_get_selected_drives_partial = partial(parallel_get_selected_drives, bucket_name=bucket_name,
                                                   latest=latest, latest_files=latest_files)

    # get the latest file for each dongle
    parallel_results = pqdm(n_jobs=num_cores, function=parallel_get_selected_drives_partial,
                            array=dongle_ids)
    
    # get indexes of files in parallel_results that have len == 0
    no_files = [ (i,x) for i,x in enumerate(parallel_results) if len(x[0])==0]
    
    # map indexes to dongle ids
    missing_dongles = [ dongle_ids[i] for i,x in no_files ]
    
    print(f"Not found any files for dongles: {missing_dongles}")

    df = pd.DataFrame(parallel_results).reset_index()
    # remap a series of rows in df[0] to a dataframe
    df = df[0].apply(pd.Series).reset_index(drop=True)
    
    df = df[~df.ext.isna()]
    return df
    

def get_drives(target, bucket_name: str = 'creationlabs-raw-data', metadata_file: str = '/home/jakub/data/asco/devices.csv',
                        show_n: int = 50, filter=0):
    # TODO: profile
    # get all dongle ids
    import humanize

    pd.set_option('display.max_colwidth',70)
    
    s3 = boto3.session.Session(region_name='eu-west-1').client('s3')
    s3_resource = boto3.resource('s3')

    if target=='-1':
        df = get_dongle_drives(s3, s3_resource, bucket_name, target, filter, latest=True)
    elif target=='0':
        df = get_dongle_drives(s3, s3_resource, bucket_name, target, filter)
    else:
        df = get_drives_df(s3_resource, bucket_name, dongle_id=target, filter=filter)

    df = df[df.drive_time.str.len() > 16 ]

    df['upload_time'] = df.drive.str.split('--').str[:-1].str.join('--')
    df['upload_time'] = pd.to_datetime(df.upload_time, format='%Y-%m-%d--%H-%M-%S')

    # get naturaltime from now to the time of the latest file
    df['time'] = df['upload_time'].apply(lambda x: humanize.naturaldelta(x))
    df = df.sort_values(by='upload_time', ascending=False)

    if target in ['0', '-1']:
        if filter: 
            df = add_metadata(df, metadata_file)
            target_columns = ['dongle_id','drive','time','seg_num','Registration_Number', 'Company']
            print(df[target_columns])
        else: print(df[['dongle_id','drive','time','seg_num']])
    else:
        df['link'] = df.apply(lambda x: urlifier(x.dongle_id, x.drive_time), axis=1)
        df = df[['drive_time','seg_num','link']]
        
        # max display size
        pd.set_option('display.max_colwidth',term_width)
        
        # drop duplicates in link and get max seg_num 
        df = df.drop_duplicates(subset=['link'], keep='last')
        df = df.sort_values(by='drive_time', ascending=False)
        
        # print the latest files
        print(df.head(show_n))

def pull_latest(dongle_id):
    """
    Download the latest file in the bucket.
    """
    bucket = 'creationlabs-raw-data'
    s3 = boto3.resource('s3')
    obj = s3.Object(bucket, dongle_id)
    files = obj.meta.client.list_objects(Bucket=bucket, Prefix=dongle_id)
    last_segment = '/'.join(files['Contents'][-1]['Key'].split('/')[:-1])
    last_drive = last_segment.split('--')[1]
    # s3.meta.client.download_file(bucket, last_drive, dongle_id)
    cmd = f'aws s3 sync s3://{bucket}/{dongle_id} .  --exclude="*" --include="{last_drive}*"'
    print(f"Running {cmd}")
    os.system(cmd)

# @profile
def add_metadata(df: pd.DataFrame, metadata_file_path: str):
    """
    Add metadata to the dataframe.
    """

    if not os.path.exists(metadata_file_path):
        print(f"Metadata file {metadata_file_path} not found")
        return df

    meta_data = pd.read_csv(metadata_file_path)
    meta_data.columns = [x.replace(' ', '_') for x in meta_data.columns]
    
    # remove all rows with na in Dongle_ID
    meta_data = meta_data[~meta_data.Dongle_ID.isna()]

    # remove values where Deprecated = True
    meta_data = meta_data[~(meta_data.Deprecated==True)]
    
    # merge the two dataframes
    df = df.merge(meta_data, left_on='dongle_id', right_on="Dongle_ID")
    
    return df


# taken from 'man ssh'

KNOWN_PARAMS = (
    "AddKeysToAgent",
    "AddressFamily",
    "BatchMode",
    "BindAddress",
    "CanonicalDomains",
    "CanonicalizeFallbackLocal",
    "CanonicalizeHostname",
    "CanonicalizeMaxDots",
    "CanonicalizePermittedCNAMEs",
    "CertificateFile",
    "ChallengeResponseAuthentication",
    "CheckHostIP",
    "Cipher",
    "Ciphers",
    "ClearAllForwardings",
    "Compression",
    "CompressionLevel",
    "ConnectionAttempts",
    "ConnectTimeout",
    "ControlMaster",
    "ControlPath",
    "ControlPersist",
    "DynamicForward",
    "EscapeChar",
    "ExitOnForwardFailure",
    "FingerprintHash",
    "ForwardAgent",
    "ForwardX11",
    "ForwardX11Timeout",
    "ForwardX11Trusted",
    "GatewayPorts",
    "GlobalKnownHostsFile",
    "GSSAPIAuthentication",
    "GSSAPIKeyExchange",
    "GSSAPIClientIdentity",
    "GSSAPIDelegateCredentials",
    "GSSAPIRenewalForcesRekey",
    "GSSAPITrustDns",
    "GSSAPIKexAlgorithms",
    "HashKnownHosts",
    "Host",
    "HostbasedAuthentication",
    "HostbasedKeyTypes",
    "HostKeyAlgorithms",
    "HostKeyAlias",
    "HostName",
    "IdentitiesOnly",
    "IdentityAgent",
    "IdentityFile",
    "Include",
    "IPQoS",
    "KbdInteractiveAuthentication",
    "KbdInteractiveDevices",
    "KexAlgorithms",
    "LocalCommand",
    "LocalForward",
    "LogLevel",
    "MACs",
    "Match",
    "NoHostAuthenticationForLocalhost",
    "NumberOfPasswordPrompts",
    "PasswordAuthentication",
    "PermitLocalCommand",
    "PKCS11Provider",
    "Port",
    "PreferredAuthentications",
    "email",
    "token",
    "Protocol",
    "ProxyCommand",
    "ProxyJump",
    "ProxyUseFdpass",
    "PubkeyAcceptedKeyTypes",
    "PubkeyAuthentication",
    "RekeyLimit",
    "RemoteForward",
    "RequestTTY",
    "RhostsRSAAuthentication",
    "RSAAuthentication",
    "SendEnv",
    "ServerAliveInterval",
    "ServerAliveCountMax",
    "StreamLocalBindMask",
    "StreamLocalBindUnlink",
    "StrictHostKeyChecking",
    "TCPKeepAlive",
    "Tunnel",
    "TunnelDevice",
    "UpdateHostKeys",
    "UsePrivilegedPort",
    "User",
    "UserKnownHostsFile",
    "VerifyHostKeyDNS",
    "VisualHostKey",
    "XAuthLocation"
)


import sys
import ipdb
import pandas as pd
import datetime as dt
import subprocess as sp
from Levenshtein import distance as dist
import os
import json
import humanize
import time

try:
    term_wdith = os.get_terminal_size().columns
except OSError:
    term_wdith = 80

pd.set_option('display.max_colwidth',term_wdith - 20)
pd.set_option('display.max_rows', None)

# TODO: treat 'cd' in a special way s.t. we can generate the absolute path wherever possible
FILTER_TERMS = ['ls','install','dt hist', 'git ', 'dt ', 'code',
                'ipython','htop','git push','git pull','git stash',
                'tmux', 'nvidia-smi', 'ping']

def dt_conv(l: str):
    try:
        pd_dt_repr = pd.to_datetime(dt.datetime.fromtimestamp(l[1]))
        return  humanize.naturaldelta(pd_dt_repr)
    except ValueError as e:
        # TODO: why does this occur
        print('ValueError', l)


def filter_term(term: tuple, excluded_terms = FILTER_TERMS):
    # Logic: all filter terms need to _not_ match
    if 'cd' in term[0][:2]:
        try:
            if term[0].split(' ')[1][0] not in ['~','/']:
                return False
            else:
                return True
        except:
            print(f"Error with {term[0]}")

    return all([ft not in term[0] for ft in excluded_terms])

def check_if_intlike(s:str):
    try:
        int(s)
        return True
    except ValueError:
        return False

def hist_tail(n_lines=20, excluded_terms = []):
    excluded_terms = FILTER_TERMS + excluded_terms
    
    shell = os.environ['SHELL'].split('/')[-1]

    if shell=='sh': shell = 'bashrc'

    # checks that we can find 
    zsh_hist = f"{os.path.expanduser('~')}/.{shell}_history"
    if not os.path.exists(zsh_hist):
        raise NotImplementedError('Not sure where to look')
    try:
        split_f = str(sp.check_output(['tail','-n',str(n_lines * 3), zsh_hist] )).split(':')
    except sp.CalledProcessError:
        sh_hist = f"{os.path.expanduser('~')}/.bash_history"
        split_f = str(sp.check_output(['tail',sh_hist,'-n',str(n_lines * 3), zsh_hist])).split(':')
        
        
    # Work Ubuntu 20.04 so far so good
    # TODO: fix this horrible line 
    parsed_commands = [ ''.join(l.replace('\\n','').strip().split(';')[-1:]) for l in split_f if '\\n' in l]
    # check if an item in split_f is convertable to int
    parsed_dt = [ int(t) for t in split_f if check_if_intlike(t)]

    parsed = list(zip(parsed_commands, parsed_dt))

    # remove FILTER_TERMS 
    cmds = [ t for t in parsed if filter_term(t, excluded_terms) ]

    # identify Levenstein distance
    lev_dist = [ dist(cmds[i][0],cmds[i+1][0]) for i in range(len(cmds)-1) ]

    # TODO: make more efficent
    # Lev dist with previous terms (1 approx to catch typos)
    ld = lev_dist + [9999]
    # dist_dict = { k:v for k,v in zip(cmds,ld) }
    dist_dict = { k:v for k,v in zip(cmds,ld) if v>4 }
    # TODO check if ordered before
    last_list = list(dist_dict.keys())[-n_lines:]

    # constructs DF with one line per command
    last_cmds = [ l[0].replace('\\','') for l in last_list ]
    # TODO: the time seems to be wrong still
    last_dt = [ dt_conv(l) for l in last_list]
    df = pd.DataFrame([last_cmds,last_dt]).T
    df.columns = ['Command', 'Time']

    return df

def hist_save(lines_to_save: list, snippet_name: str, n_lines: str):
    df = hist_tail(n_lines)
    
    snippets_path = os.path.join(os.path.expanduser('~'), 'dt_config.json')
    config = json.loads(open(snippets_path).read())

    # prepare the correct saving format
    saved_time = dt.datetime.now().strftime('%Y-%m-%d %H-%M-%S')
    # TODO: figure out why there's this horrible off by-one error
    parsed_lines = [df.iloc[x-1].Command for x in lines_to_save]
    saved_lines = [snippet_name, saved_time, parsed_lines]
    
    config['snippets'].append(saved_lines)
    
    json.dump(config, open(snippets_path, 'w'))
    
    


import os
import json

def get_sftp_path():
    '''
    Checks if SFTP config exists.
    '''
    cwd = os.getcwd()
    target_path = os.path.join(cwd, '.vscode/sftp.json')
    return target_path

def write_sftp(name: str, remote: str, ip: int,
                user: str, port: int, key_path: str):
    sftp_path = get_sftp_path()
    new_config = {
        "name": name,
        "host": ip,
        "protocol": "sftp",
        "port": port,
        "username": user,
        "remotePath": remote,
        "privateKeyPath" : key_path,
        "uploadOnSave": True,
        "watcher": {
            "files": "*.{py,.sh,.txt}"
        },
        "ignore": [
            ".vscode",
            "*.png",
            "*.jpg",
            "*.npy",
            ".git/", 
            "*.pt",
            "wandb/"
        ]
    }
    if not os.path.exists(sftp_path):
        os.mkdir(os.path.join(os.getcwd(), '.vscode'))
        # create sftp
        with open(sftp_path, 'w') as fp:
            fp.write(json.dumps(new_config).replace("'", ""))
        print(new_config)

    else:
        old_config = json.load(open(sftp_path, 'r'))
        # TODO: update values from new config
        config = {
            **new_config
        }
        with open(sftp_path, 'w') as fp:
            fp.write(json.dumps(new_config).replace("'", ""))
        print(new_config)
        



from dt.ext.aws_s3 import get_lastest_file
from cement import Controller, ex
from cement.utils.version import get_version_banner
from ..core.version import get_version
from pprint import pprint as pp
import os

VERSION_BANNER = """
ML & data helper code! %s
%s
""" % (get_version(), get_version_banner())


class Base(Controller):
    class Meta:
        label = 'base'

        # text displayed at the top of --help output
        description = 'ML & data helper code!'

        # text displayed at the bottom of --help output
        epilog = 'Usage: dt command [args] [kwargs]'

        # controller level arguments. ex: 'dt --version'
        arguments = [
            ### add a version banner
            ( [ '-v', '--version' ],
              { 'action'  : 'version',
                'version' : VERSION_BANNER } ),
        ]


    def _default(self):
        """Default action if no sub-command is passed."""

        self.app.args.print_help()

    @ex(
        # TODO: implement `--home` for setting up home address
        help='AWS Security Groups helper functions',
        arguments=[
            (['action'],{
                "help" : "One of {update, show, ls}",
                "action": "store"
            }),
            (['-n','--name'],{
                "help" : "Name of the SG to act on",
                "action": "store"
            }),
            (['-ip'],{
                "help" : "IP to add to security group",
                "action": "store"
            }),
            (['-p','--profile'],{
                "help" : "Which AWS profile should boto3 use",
                "action" : "store"
            }),
            (['-P','--ports'],{
                "help" : "Which ports should AWS SG set",
                "action" : "store"
            }),
            (['-m','--messages'],{
                "help" : "Message to add to SG description, WARNing will not use default ports if used "
                         "Accepts port:messages pairs separated by commas. E.g. 42:answer to life, 22:memes",
                "action" : "store"
            }),
            (['-r','--region'],{
                "help": "Which AWS region should boto3 use",
                "action": "store"
            }),
            (['-i','--id'],{
                "help" : "ID of the SG to act on from the ls DataFrame",
                "action" : "store"
            }),
        ]
    )
    def sg(self):
        from ..ext.aws_sg import get_sg, update_sg_to_ip
        profile = self.app.pargs.__dict__.get('profile') or self.app.pargs.__dict__.get('p') or 'default'
        name = self.app.pargs.__dict__.get('name') or self.app.pargs.__dict__.get('n')
        region = self.app.pargs.__dict__.get('region') or self.app.pargs.__dict__.get('r') or 'eu-west-1'

        # which ports to change
        ports = self.app.pargs.__dict__.get('ports') or self.app.pargs.__dict__.get('P')
        if ports: ports = [ int(p) for p in ports.split(',') ]
        else: ports = [22, 8888, 8501, 8889, 5432]

        # which messages to assign to which ports
        messages = self.app.pargs.__dict__.get('messages') or self.app.pargs.__dict__.get('m') or ''

        # if any id is given, use that one
        sid = self.app.pargs.__dict__.get('id') or self.app.pargs.__dict__.get('i')
        if sid: sid = int(sid)

        if self.app.pargs.action=='show':
            pp(get_sg(profile, sid = sid, region=region))

        elif self.app.pargs.action=='update':
            if not self.app.pargs.name: sg_name = get_sg(profile, sid = sid)[0]['GroupName']
            else: sg_name = self.app.pargs.name
            update_sg_to_ip(sg_name=sg_name, profile=profile,
                            ports=ports, messages=messages, region=region)

        elif self.app.pargs.action=='ls':
            from ..ext.aws_sg import list_security_groups
            df = list_security_groups(region, profile)
            print(df)

        else:
            raise NotImplementedError('As action, chose one of {update, show}')

    @ex(
        help="GCP VM Instances helper commands."
        "These are currently in alpha and some may not work",
        arguments=[
            (['action'],{
                    "help" : "To show one of {ls, scp}.",
                    "action" : "store"
                }),
            (['--rpath'],{
                "help" : "What is the location on the remote server",
                "action": "store"
            }),
            (['-f','--filter'],{
                "help" : "One of {live, all}. Default: live",
                "action": "store"
            }),
            (['-id'],{
                "help": "InstanceId to be passed to ec2 start/stop",
                "action": "store"
            })
        ]
    )
    def gcp(self):
        from ..ext.gcp import gcp_ls, scp

        # profile = self.app.pargs.profile or 'default'
        # filt = self.app.pargs.filter or "live"

        if self.app.pargs.action=='ls':
            print(gcp_ls())

        # elif self.app.pargs.action=='update':
        #     print(update_ec2_ssh(profile))

        # elif self.app.pargs.action=='start':
        #     start_ec2(self.app.pargs.id, profile)

        elif self.app.pargs.action=='scp':
            scp(int(self.app.pargs.id), self.app.pargs.rpath)
            # stop_ec2(self.app.pargs.id, profile)

        else:
            raise NotImplementedError("Choose one of {ls, update, start, stop}")

    @ex(
        help='Tree helper/replacement functions',
        arguments= [
            (['loc'], {
                "help": "Location of the tree to act on",
                "action": "store"
            }),
            (['-n'],{
                "help": "Depth of the tree to print",
                "action": "store"
            }),
            (['-f'],{
                "help": "Print out full path",
                "action": "store"  
            })
        ]
    )
    def tree(self):
        from ..ext.tree import tree
        
        level = int(self.app.pargs.n or 2)
        full_path = int(self.app.pargs.f or 0)
        
        tree(self.app.pargs.loc, level, full_path)
        

    @ex(
        help="AWS EC2 helper commands.",
        arguments=[
            (['action'],{
                    "help" : "To show one of {ls, update}.",
                    "action" : "store"
                }),
            (['-p','--profile'],{
                "help" : "Which AWS profile should boto3 use.",
                "action": "store"
            }),
            (['-f','--filter'],{
                "help" : "One of {live, all}. Default: live",
                "action": "store"
            }),
            (['-id'],{
                "help": "InstanceId to be passed to ec2 start/stop",
                "action": "store"
            }),
            (['-r','--region'],{
                "help": "Which region to use",
                "action": "store"
            }),
        ]
    )
    def ec2(self):
        from ..ext.aws_ec2 import get_ec2, start_ec2, stop_ec2, update_ec2_ssh

        profile = self.app.pargs.profile or 'default'
        filt = self.app.pargs.filter or "live"
        region = self.app.pargs.region or "eu-west-1"

        if self.app.pargs.action=='ls':
            print(get_ec2('live', profile, filt, region))

        elif self.app.pargs.action=='update':
            print(update_ec2_ssh(profile))

        elif self.app.pargs.action=='start':
            start_ec2(self.app.pargs.id, profile)

        elif self.app.pargs.action=='stop':
            stop_ec2(self.app.pargs.id, profile)

        else:
            raise NotImplementedError("Choose one of {ls, update, start, stop}")

    @ex(
        help="Operations to easily work with remote servers / devices",
        arguments=[
            # TODO: maybe make grabable from ec2 ls
            (['ip'],{
                "help" : "IP of the target server",
                "action" : "store"
            }),
            (['-r','--remote'],{
                "help" : "Remote path on the target server",
                "action": "store"
            }),
            (['-u','--user'],{
                "help" : "Which user to ssh with.",
                "action" : "store"
            }),
            (['--server_name','-sn'],{
                "help" : "To give the SFTP server a name.",
                "action" : "store"
            }),
            (['-i', '--identity'],{
                "help" : "Path to the identity file to use for ssh.",
                "action" : "store"
            }),
            (['-p','--port'],{
                "help" : "Port over which to run SFTP.",
                "action" : "store"
            })
        ]
    )
    def sftp(self):
        ip = self.app.pargs.ip
        remote = self.app.pargs.__dict__.get('remote') or self.app.pargs.__dict__.get('r') or '/data/openpilot'
        port = self.app.pargs.__dict__.get('port') or self.app.pargs.__dict__.get('p') or 8022
        user = self.app.pargs.__dict__.get('user') or self.app.pargs.__dict__.get('u') or 'comma'
        server_name = self.app.pargs.__dict__.get('server_name') or self.app.pargs.__dict__.get('sn') or 'C3'
        identity = self.app.pargs.__dict__.get('identity') or self.app.pargs.__dict__.get('i') or os.path.expanduser("~/openpilot/tools/ssh/CL-key")

        from ..ext.sftp_conf import write_sftp
        
        write_sftp(name=server_name, ip=ip, port=port, user=user,
                    remote=remote, key_path=identity)

    @ex(
        help="Ops with Codex AI from OpenAI.",
        arguments=[
            (['text'],{
                "help" : "Text to be suggested",
                "action": "store"
            }),
            (['-n','--n_suggestions'],{
                "help" : "Number of suggestions to return",
                "action": "store",
                "default": 5
            })
        ]
    )
    def cai(self):
        from ..ext.codex import get_cai_suggestions
        import pprint as pp

        suggestions = get_cai_suggestions(self.app.pargs.text, self.app.pargs.n_suggestions)
        pp.pprint(suggestions)


    @ex(
        help='Monitor lack of GPU activity. When there is none, runs -job',
        arguments=[
            (['-job'],{
                "help" : "Text of the command to run after no GPU activity",
                "action" : "store"
            }),
            (['-n'], {
                "help" : "Number of processes before 'no activity'.",
                "action" : "store"
            })
        ]
    )
    def monitor(self):
        from ..ext.monitor import submit
        submit(self.app.pargs.job)


    # number of files
    @ex(
        help='Displays the number of files',
        arguments=[
            (['-d'], {
            "help" : "which directory to show",
            "action" : "store" } )
        ]
    )
    def nf(self):
        loc = self.app.pargs.d or '.'
        os.system(f'ls {loc} | wc -l')

    @ex(
        help="Appends defaults: from zshrc.txt to ~/.bashrc or ~/.zshrc or tmux conf",
        arguments = [
            (['shell'], {
                "help" : "One of {sh, zsh, tmux}. Imports a DT version of zshrc into bashrc or zshrc. or tmux .conf into ~/",
                "action": "store"
            })
        ]
    )
    def load(self):
        # TODO: reload the file after.
        # TODO: load git aliases
        zshrc = os.path.abspath(__file__).split('/')[:-2] + ['ext', 'zshrc.txt']
        tmux = os.path.abspath(__file__).split('/')[:-2] + ['ext', 'tmux.conf']

        zshrc_path = os.path.join('/'.join(zshrc))
        tmux_path = os.path.join('/'.join(tmux))
        if self.app.pargs.shell == 'sh':
            os.system(f'cat {zshrc_path} >> ~/.bashrc ')
        elif self.app.pargs.shell == 'zsh':
            os.system(f'cat {zshrc_path} >> ~/.zshrc ')
        elif self.app.pargs.shell == 'tmux':
            os.system(f'cat {tmux_path} >> ~/.tmux.conf ')
        else:
            raise NotImplementedError("Try one of {sh, zsh, tmux}")


    @ex(
        help='Operations on CLI hist',
        arguments=[
            (['-s'],{
                "help" : "Which lines to save separated by commas",
                "action" : "store"
            }),
            (['-n'],{
                # TODO: probably better as mandatory
                "help" : "Which lines from the selected history to show/save",
                "action" : "store"
            }),
            (['-e'],{
                "help": "Which terns to exclude separated by commas",
                "action": "store"
            }),
            (['-a','--alias'],{
                "help": "Name for the snippet.",
                "action" : "store"
            })
        ]
    )
    def hist(self):
        if not self.app.pargs.s:
            from ..ext.hist import hist_tail
            import pprint as pp
            n_lines = int(self.app.pargs.n or 20)
            excluded_terms = self.app.pargs.e or []
            if isinstance(excluded_terms, str): excluded_terms = excluded_terms.split(',')
            pp.pprint(hist_tail(n_lines, excluded_terms))
        else:
            from ..ext.hist import hist_save
            import uuid
            
            n_lines = int(self.app.pargs.n or 20)
            lines_to_save = [int(x) for x in str(self.app.pargs.s).split(',')]
            snippet_name = self.app.pargs.__dict__.get('a') or self.app.pargs.__dict__.get('alias') or str(uuid.uuid4())
            print(f'Saving {lines_to_save} lines from history to snippet {snippet_name}')
            
            hist_save(lines_to_save, snippet_name, n_lines=n_lines)

    @ex(
        help='Operations with snippets',
        arguments=[
            (['-n','--n_lines'],{
                "help" : "Which snippet to run/create. Using commas to seperate ids"
                "e.g. `dt run 13,23`",
                "action" : "store"
            }),
            (['-ssh'],{
                "help" : "Whether the command (includes?)/is? an SSH command",
                "action": "store"
            }),
            (['-c'],{
                # TODO: probably better as mandatory
                "help" : "Which lines from the selected history to show/save",
                "action" : "store"
            }),
            # TODO: what if non-default n lines has been pased to hist
            (['--save'],{
                "help" : "Whether to save the snippet to the history",
                # TODO: figure out if bool
                "action" : "store"   
            }),
            (['-a','--alias'],{
                "help": "Name of the alias to run",
                "action" : "store"
            })
        ]
    )
    def run(self):
        is_alias = self.app.pargs.__dict__.get('a') or self.app.pargs.__dict__.get('alias') or False
        
        if is_alias:
            from ..ext.run import run_command_from_alias
            run_command_from_alias(is_alias)
        else:
            from ..ext.run import run_command
            n_lines = [int(x) for x in str(self.app.pargs.n_lines).split(',')]
            run_command(n_lines)
            
    @ex(
        help='Operations with s3 bucket creationlabs-raw-data',
        arguments=[
            (['action'],{
                "help" : "One of {ls, link}",
                "action" : "store"
            }),
            (['target'],{
                "help" : "The target of the action. Can be {dongle_id, -1}",
                "action" : "store"
            }),
            (['--num_segments'],{
                "help" : "Number of recent files to show",
                "action" : "store",
                "default" : 10
            }),
            (['-n','--show_n'],{
                "help" : "Number of recent files to show",
                "action" : "store",
                "default" : 50
            }),
            (['-f','--filter'],{
                "help" : "Filter the files by this int",
                "action" : "store",
                "default" : 0
            }),
            (['-F','--file'],{
                "help": "File containing metadata",
                "action": "store",
                "default": "/home/jakub/data/devices.csv"
            })
        ]
    )
    def s3(self):
        if self.app.pargs.action == 'link':
            from ..ext.aws_s3 import get_lastest_file
            print(get_lastest_file(self.app.pargs.target))

        # ls actions for all dongle ids
        if self.app.pargs.action == 'ls':
            filter = self.app.pargs.__dict__.get('filter') or self.app.pargs.__dict__.get('f') or 0
            metadata_file = self.app.pargs.__dict__.get('file') or self.app.pargs.__dict__.get('F')
            
            from ..ext.aws_s3 import get_drives
            show_n = int(self.app.pargs.__dict__.get('show_n') or self.app.pargs.__dict__.get('n'))

            if self.app.pargs.target == '-1':
                get_drives(self.app.pargs.target, show_n=show_n, metadata_file=metadata_file, filter=filter)
            else:
                # TODO: potentially removed
                # num_segments = self.app.pargs.__dict__.get('n') or self.app.pargs.__dict__.get('num_segments')
                get_drives(self.app.pargs.target, show_n=show_n, metadata_file=metadata_file, 
                            filter=int(filter))

        if self.app.pargs.action == 'pull':
            from ..ext.aws_s3 import pull_latest
            pull_latest(self.app.pargs.target)

    @ex(
        help='Operations on the SSH config',
        arguments=[
            (['action'],{
                "help" : "To show one of {set, ls, update}. Set is for new remote value, update is to grab from AWS",
                "action" : "store"
            }),
            (['-host'],{
            "help" : "Which SSH config host to SET",
            "action": "store"
            }),
            (['-ip'],{
                "help" : "The IP to SET in the ssh config as HostName",
                "action" : "store"
            })
        ]
    )
    def ssh(self):
        from ..ext.sshconf import read_ssh_config
        from ..ext.sshconf import print_ls
        ssh_config_path = os.path.expanduser('~/.ssh/config')

        if self.app.pargs.action=='set':
            ssh_config_obj = read_ssh_config(ssh_config_path)
            ssh_config_obj.set(self.app.pargs.host, HostName=self.app.pargs.ip)
            ssh_config_obj.write(ssh_config_path)
            print_ls(ssh_config_obj)
            

        elif self.app.pargs.action=='update':
            from ..ext.aws_ec2 import update_ec2_ssh
            profile ='default' # or  self.app.pargs
            # print(update_ec2_ssh(profile))
            update_ec2_ssh(profile)
            ssh_config_obj = read_ssh_config(ssh_config_path)
            print_ls(ssh_config_obj)

        elif self.app.pargs.action=='ls':
            ssh_config_obj = read_ssh_config(ssh_config_path)
            print_ls(ssh_config_obj)

        else:
            # raise NotImplementedError()
            print('Arg out of {set, update} provided assuming an SSH command.')
            ssh_config_obj = read_ssh_config(ssh_config_path)
            target = self.app.pargs.action
            if not target in ssh_config_obj.hosts_: raise Exception(f'Host {target} not in configured hosts.')
            else:
                os.system('dt sg update')
                os.system(f'ssh {target}')

    @ex(
        help='Basic viz using streamlit for image comparisons',
        arguments = [
            (['images'], {
                'help' : 'Location of target images',
                'action' : "store"
            }),
            (['labels'], {
                'help' : 'Location of target labels',
                "action" : 'store'
            }),
            (['adapted'], {
                'help' : 'Location of domain adapted images',
                "action" : "store"
            })
            #     'action' : "store",
            #     "images" : '/efs/public_data/gta5/images',
            #     "labels" : '/efs/public_data/gta5/labels',
            #     "adapted" : '/efs/public_data/images'
            # })
        ]
    )
    def viz(self):
        args = self.app.pargs
        print(args.images)
        from ..ext.firelit import create_viz
        create_viz(args.images, args.labels, args.adapted)

    @ex(
        help="Download file from Google Drive",
        arguments=[
            (["file_id"], {
                "help" : "The file id of the file to download",
                "action" : "store"
            }),
            (["destination"], {
                "help": "The destination to download the file to",
                "action" : "store"
            })
        ]
    )
    def gdrive(self):
        from ..ext.gdrive_dl import download_file_from_google_drive
        download_file_from_google_drive(self.app.pargs.file_id, self.app.pargs.destination)


    @ex(
        help='Operations with config.',
        arguments = [
            (['action'], {
                "help" : "One of {show, set, loc}: \
                *show*: show configs Sentry DNS etc. \
                *set*: Resets the config from CLI. \
                *loc*: shows the location of the config",
                'action' : "store",
            })
        ]
    )
    def config(self):
        # TODO: setup Data-toolkit specific config
        # shows config
        
        if self.app.pargs.action=='show' or self.app.pargs.action=='ls':
            from ..ext.config import show_config
            show_config()

        # shows the path of the config
        if self.app.pargs.action=='loc':
            config_path = os.path.join(os.path.expanduser('~'), 'dt_config.json')
            print(config_path)

        # resets the config
        # TODO this config method needs programmatic acces for spot instances
        if self.app.pargs.action=='set':
            from ..ext.tracking import initialize
            initialize(overwrite=True)

        elif self.app.pargs.action not in ['set','loc', 'show','ls']:
            raise NotImplementedError("Try one of {show, set, loc}")


    @ex(
        help='All Github related operations',
        arguments=[
             (['action'],{
                "help" : """To show one of {ff, fs, contrib, diff}. \r\n
                ff: Finds the forks of a repo.  \r\n
                fs: Finds the last commit of a repo by fork. \r\n 
                contrib: Finds the contributors of a repo. \r\n
                diff: Finds the all the diffs in directories below. """,
                "action" : "store"
            }),
            (['-url'],{
                # TODO: what if there's no forks to give?
                "help" : "Github URL to search for forks."
            })
        ]
    )
    def gh(self):
        if self.app.pargs.action=='ff':
            from ..ext.github import find_forks
            print(find_forks(self.app.pargs.url))
        elif self.app.pargs.action=='fs':
            from ..ext.github import fork_status
            print(fork_status(self.app.pargs.url))
        elif self.app.pargs.action=='contrib':
            from ..ext.github import find_contributors
            print(find_contributors(self.app.pargs.url))
        elif self.app.pargs.action=='diff':
            from ..ext.github import dir_diff
            pp(dir_diff(self.app.pargs.url))


    @ex(
        help='Manage TODOs using Google Keep',
        arguments=[
            (['action'],{
                "help" : "One of {add, ls, dn}",
                "action": "store"
            }),
            (['-t'],{
                "help": "Text of the TODO",
                "action": "store"
            }),
            # (['-l'],{
            #     "help": "Length of time in hours this is expected to take",
            #     "action": "store"
            # }),
            (['-p'],{
                "help" : "Project id. Daily is 0",
                "action" : 'store'
            }),
            (['-i'],{
                "help" : "Id of the finished task",
                "action": "store"
            }),
            (['-f'],{
                "help" : "Filter TODOs for ls",
                "action": "store"
            })
        ]
    )
    def td(self):
        from ..ext.asana import add_todo, asana_list_todos, done_todo
        project = int(self.app.pargs.p or 0)

        if self.app.pargs.action=='add':
            add_todo(self.app.pargs.t, project)
        elif self.app.pargs.action=='ls':
            asana_list_todos(self.app.pargs.f or True,
                       self.app.pargs.p)
        elif self.app.pargs.action=='dn':
            done_todo(self.app.pargs.i)

        else:
            raise NotImplemented("Action must be one of {add, ls, dn}")

    @ex(
        help='Manage TODOs using Google Keep',
        arguments=[
            (['action'],{
                "help" : "One of {add, ls, dn}",
                "action": "store"
            }),
            (['-t'],{
                "help": "Text of the TODO",
                "action": "store"
            }),
            (['-l'],{
                "help": "Length of time in hours this is expected to take",
                "action": "store"
            }),
            (['-p'],{
                "help" : "Project name. One of {DA, SL}",
                "action" : 'store'
            }),
            (['-i'],{
                "help" : "Id of the finished task",
                "action": "store"
            }),
            (['-f'],{
                "help" : "Filter TODOs for ls",
                "action": "store"
            })
        ]
    )
    def gkeep(self):
        from ..ext.gkeep import add_todo, list_todos, done_todo
        project = self.app.pargs.p or 'SL'

        if self.app.pargs.action=='add':
            add_todo(self.app.pargs.t, self.app.pargs.l, self.app.pargs.p)
        elif self.app.pargs.action=='ls':
            list_todos(self.app.pargs.f or True,
                       self.app.pargs.p)
        elif self.app.pargs.action=='dn':
            done_todo(self.app.pargs.i)

        else:
            raise NotImplemented("Action must be one of {add, ls, dn}")
        
    # TODO: implement this as command
    # likely easiest as just alias dt='python -m dt.main'
    # echo "alias dt='python -m dt.main'" > ~/.bashrc; source ~/.basshrc
    # local / mac version and EC2 / remote version

    @ex(
        help='(WIP/Pre-Alpha) Execute a Python command across all files in current dir.',
        arguments=[
            (['command'],{
                "help": "what python command to run",
                "action": "store"
            }),
            (['-f'],{
                "help" : "filter all files in in this directory and below by this expression.",
                "action" : "store"
            })
        ]
    )
    def py(self):
        from ..ext.python_exec import execute
        execute(self.app.pargs.command, self.app.pargs.f)